import json, pathlib, logging
from os import path

my_logger = logging.getLogger("ams_reg_calc")

class RegisterCalculator:
    database = list()
   
    @classmethod    
    def SetDatabase(self, database_path: str):
        """
        This method imports the database that descripbes the registers

        Args:
            database_path (str): path to the .json file that describes the registers
        """
        with open(database_path, "r") as database_file:
            self.database = json.load(database_file)

    @classmethod
    def _map_bitfield_into_reg_value(self, register_value: int, bitfield_value: int, start: int, stop: int):
        #my_logger.debug(start)
        #my_logger.debug(stop)
        mask = int("0" * (32 - stop) + f"{bitfield_value:032b}"[::-1][0:stop - start + 1][::-1] + "0" * start, 2)
        #my_logger.debug(f"{mask:032b}")
        return register_value | mask


    @classmethod
    def _mask_out_bitfield_value(self,register_value: int, start: int, stop: int) -> int:
        """
        Updates the value of a bitfield into an actually available register value

        Args:
            register_value (int): initial register value
            start (int): start position of bitfield
            stop (int): stop position of bitfield

        Returns:
            int: updated register value
        """
        my_logger.debug("start: " + str(start))
        stop = stop
        my_logger.debug("stop : " + str(stop))
        return int(f"{register_value:032b}"[::-1][start:stop + 1][::-1], 2)


    @classmethod
    def _get_register_dict_by_bitfield(self, bitfield_name: str) -> dict:
        """
        Gets the descriptor dictionary from a register database for a given bitfield

        Args:
            bitfield_name (str): name of the bitfield

        Returns:
            dict: descriptor dictionary
        """
        for act_reg in self.database:
            for act_bf in act_reg["fields"]:
                if bitfield_name.upper().strip() == act_bf["name"].upper().strip():
                    return act_reg


    @classmethod
    def _get_reg_dict_by_name(self, register_name: str) -> dict:
        """
        Gets the descriptor dictionary for a given register name

        Args:
            register_name (str): name of the register

        Returns:
            dict: descriptor dictionary
        """
        for act_reg in self.database:
            if register_name.upper().strip() == act_reg["register_name"].upper().strip():
                return act_reg


    @classmethod
    def _get_reg_dict_by_address(self, register_address: str) -> dict:
        """
        Gets the descriptor dictionary for a given register address

        Args:
            register_address (str): address of the register

        Returns:
            dict: descriptor dictionary
        """
        for act_reg in self.database:
            if int(register_address, 16) == int(act_reg["address"], 16):
                return act_reg


    @classmethod
    def EncodeRegisters(self, bitfields_dict: dict) -> dict:
        """
        This method calculates the value that has to be set based on the bitfields given to it

        Args:
            bitfields_dict (dict): dictinary that defines the bitfields (as keys) and their value

        Returns:
            dict: dictionary where the keys are the addresses of the registers and the value of the register is calculated based on the given bitfields
        """
        bitfields_dict = self._pre_correct_values(bitfields_dict)
        return self._build_registers_from_desciptors_and_dict(
            bitfields_dict,
            self._collect_register_descriptors(bitfields_dict)
        )

    @classmethod
    def _collect_register_descriptors(self, bitfields_dict: dict) -> dict:
        registers_dict = []
        for act_bf in list(bitfields_dict.keys()):
            regdict = self._get_register_dict_by_bitfield(act_bf)
            if regdict not in registers_dict:
                registers_dict.append(regdict)
        return registers_dict

    @classmethod
    def _build_registers_from_desciptors_and_dict(self, bitfields_dict: dict, registers_dict: dict) -> dict:
        return_dict = {}
        for act_reg in list(registers_dict):
            reg_value = 0x00000000
            reg_address = act_reg["address"]
            for act_bf in act_reg["fields"]:
                if act_bf["name"] in bitfields_dict.keys():
                    bf_value = bitfields_dict[act_bf["name"]]
                else:
                    conv_base = 2 if "x" in act_bf["default"] else 10
                    bf_value = int(act_bf["default"], conv_base)
                my_logger.debug(f"{act_bf['name']}: {bf_value}")
                reg_value = self._map_bitfield_into_reg_value(
                    reg_value, 
                    bf_value,
                    int(act_bf["pos"].split(":")[-1]),
                    int(act_bf["pos"].split(":")[0]),
                )
                #my_logger.debug(f"{reg_value:032b}")
            return_dict[reg_address] = reg_value

        return return_dict
            


    @classmethod
    def DecodeRegisters(self, registers_dict: dict) -> dict:
        """
        This method calculates the bitfield values of all registers given in the registes dict

        Args:
            registers_dict (dict): dictionary where the keys are the register addresses and values are the numeric values of the registers

        Returns:
            dict: dictionary where the keys are the bitfields and the values are calculated based on the given register values
        """
        registers_dict = self._pre_correct_values(registers_dict)
        return_dict = {}
        for act_reg in list(registers_dict.keys()):
            return_dict[act_reg] = self.DecodeRegister(registers_dict[act_reg], register_address=act_reg)
        return return_dict


    @classmethod
    def DecodeRegister(self, register_value: int, register_name: str=None, register_address: str=None) -> dict:
        """
        Retursn the dictionary of bitfields belonging to the register defined by either the name or the address of the register.

        Args:
            register_value (int): value of the register
            register_name (str, optional): Name of the register. Defaults to None.
            register_address (str, optional): address of the register. Defaults to int=None.

        Returns:
            dict: dictionary of the bitfields calculated from the rgister value
        """
        

        if register_address != None:
            register_dict = self._get_reg_dict_by_address(register_address)
        if register_name != None:
            register_dict = self._get_reg_dict_by_name(register_name)
        
        return_dict = {}
        for act_bf in register_dict["fields"]:
            return_dict[act_bf["name"]] = self._mask_out_bitfield_value(
                register_value,
                int(act_bf["pos"].split(":")[-1]),
                int(act_bf["pos"].split(":")[0])
            )
        return return_dict


    @classmethod
    def get_name_by_address(self, register_address: str) -> str:
        """
        This method returns the name of the register defined by the address

        Args:
            register_address (str): address of the register

        Returns:
            str: name of the register belongs to the address
        """
        for act_reg in self.database:
            if int(act_reg["address"], 16) == int(register_address, 16):
                return act_reg["register_name"]


    @classmethod
    def get_address_by_name(self, register_name: str) -> str:
        """
        Returns the address of the register defined by the name

        Args:
            register_name (str): name of the regsiter

        Returns:
            str: address of the register
        """
        for act_reg in self.database:
            if act_reg["register_name"].upper().strip() == register_name.upper().strip():
                return act_reg["address"]

    @classmethod
    def _pre_correct_values(self, input_dict: dict) -> dict:
        """
        This method is to remove mistypings from the dictionary keys.
        Mistakes:
            writespaces at the beginning or end of the keys

        Args:
            input_dict (dict): parameter dictionary

        Returns:
            dict: corrected parameter dictionary
        """
        right_dict = {}
        for act_element in list(input_dict.keys()):
            right_dict[act_element.strip()] = input_dict[act_element]

        return right_dict